/***************************************************************************
	Routines for supporting the launch of a GAMMA job on the cluster.
	Read COPYRIGHT file in this directory
****************************************************************************/

/*
 *  Try to create a virtual GAMMA across the network, assigning it a unique
 *  par_pid.
 *  On success it returns the par_pid, and a list of indexes of the nodes
 *  involved in the virtual GAMMA as its 2th arg.
 *  NOTE:  the arguments "user_space_node_list_***" point to user space!!!
 */
int
sys_gamma_vm_create
(
	unsigned long num_nodes,
	void *user_space_node_list_in,
	void *user_space_node_list_out
)
{

	int found, j, procnum;
	unsigned long i,par_pid;
	par_entry *p;
	port_entry *pp;


	if ((num_nodes < 1) || (num_nodes > MAX_NUM_NODES))
		num_nodes = total_num_nodes;

/*  choose a par_pid, if one is available  */
/*  par_pid #0 is reserved!  */
	found = 0;
	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++)
		if (
			(par_proc_table[par_pid] == (par_entry *)NULL)
			&&
			(!((par_pid-1-phys_whoami) % total_num_nodes))
		) {
			found = 1;
			break;
		}
	if (!found) 
		return -GAMMA_ERR_GENERIC;

/*  Allocate a descriptor for the new virtual GAMMA  */
	par_proc_table[par_pid] =
		(par_entry *)__get_free_pages(GFP_KERNEL,par_entry_magnitude);
	p = par_proc_table[par_pid];
	if (p == (par_entry *)NULL) {
printk("sys_gamma_vm_create(): kmalloc() failed\n");
		return -GAMMA_ERR_GENERIC;
	}

	for (i=0;i<MAX_NUM_NODES;i++)
		p->vm.wish_flag[i] = 0;

/*  The invoking process is always instance #0 and procnum #0  */
	p->inst_table[0].pid = -1;
	p->inst_table[0].task = NULL;
	p->inst_table[0].whoami = 0;
	p->inst_table[0].has_pkts_in_rx_tmp_ring = 0;

/*  No other local instances yet  */
	for (i=1;i<MAX_NUM_LOCAL_INST;i++) {
		p->inst_table[i].pid = -1;
		p->inst_table[i].task = NULL;
		p->inst_table[i].whoami = -1;
	}

	procnum = 0;

	if (user_space_node_list_in == NULL) {

/*
 *  Choose "num_nodes" nodes to create a virtual GAMMA,
 *  indexing them from 0 on, fill in the virtual GAMMA
 *  descriptor, find out how many process instances are allocated on the
 *  local host ("procnum"), and allocate room for pre-computed pkt headers.
 *  Nodes are chosen circularly in file /etc/gamma.conf, starting
 *  from local host.
 */

		p->vm.how_many_nodes = num_nodes;

		for (i=0;i<num_nodes;i++) {
			p->eth_table[i] =
				eth_table[(phys_whoami+i) % total_num_nodes];
			p->vm.node_list[i] = (phys_whoami+i) % total_num_nodes;
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printk("sys_gamma_vm_create():  too many proc instances on this node.  Aborting...\n");
/*
 *  Virtual GAMMA not yet created across the network.  Prevent DESTROY
 *  message from being transmitted by sys_gamma_vm_destroy()...
 */
				p->vm.how_many_nodes = 1;
/*  Destroy the virtual GAMMA  */
				sys_gamma_vm_destroy(par_pid);
				return -GAMMA_ERR_GENERIC;
			}
			p->inst_table[procnum-1].whoami = i;
			p->inst_table[procnum-1].has_pkts_in_rx_tmp_ring = 0;
			for (j=0;j<=MAX_PORT_NUMBER;j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->task = NULL;
				pp->has_rec_handler =
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_same_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = (l2s *)kmalloc(
					HEADER_SIZE+MAX_1ST_PIECE_SIZE,
					GFP_KERNEL
				);
				if (pp->header == NULL) {
printk("sys_gamma_vm_create():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
			}
		}

	}

	else {

/*
 *  Take "num_nodes" nodes from the user-provided list of machines,
 *  then build the virtual GAMMA similar to the above "if" branch.
 */
		memcpy_fromfs(
			p->vm.node_list,user_space_node_list_in,
			num_nodes*sizeof(unsigned char)
		);
		p->vm.node_list[0] = phys_whoami;
		p->vm.how_many_nodes = num_nodes;

		for (i=0;i<num_nodes;i++) {
			p->eth_table[i] = eth_table[p->vm.node_list[i]];
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printk("sys_gamma_vm_create():  too many proc instances on this node.  Aborting...\n");
/*
 *  Virtual GAMMA not yet created across the network.  Prevent DESTROY
 *  message from being transmitted by sys_gamma_vm_destroy()...
 */
				p->vm.how_many_nodes = 1;
/*  Destroy the virtual GAMMA  */
				sys_gamma_vm_destroy(par_pid);
				return -GAMMA_ERR_GENERIC;
			}
			p->inst_table[procnum-1].whoami = i;
			p->inst_table[procnum-1].has_pkts_in_rx_tmp_ring = 0;
			for (j=0;j<=MAX_PORT_NUMBER;j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->task = NULL;
				pp->has_rec_handler =
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_same_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = (l2s *)kmalloc(
					HEADER_SIZE+MAX_1ST_PIECE_SIZE,
					GFP_KERNEL
				);
				if (pp->header == NULL) {
printk("sys_gamma_vm_create():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
			}
		}

	}

	p->num_locinst = procnum;

/*  Return the chosen nodes to the caller in user space  */
	memcpy_tofs(
		user_space_node_list_out,p->vm.node_list,
		num_nodes*sizeof(unsigned char)
	);

/*  Register the virtual GAMMA across the network, via broadcast  */
	gamma_kernel_send(
		BROADCAST,par_pid,VM_CREATE,
		(void *)(&(p->vm)),sizeof(vm_descriptor),1
	);
/*  Wait enough (1 msec) for the registration to occur remotely  */
	busywait(1000);

	return par_pid;

}


int
sys_gamma_vm_create_pid
(unsigned long par_pid, unsigned long num_nodes, void *user_space_node_list)
{

	unsigned long i;
	int j,procnum;
	par_entry *p;
	port_entry *pp;


	if ((num_nodes < 1) || (num_nodes > MAX_NUM_NODES))
		num_nodes = total_num_nodes;

/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	p = par_proc_table[par_pid];

/*  A virtual GAMMA should already exist!  */
	if (p == (par_entry *)NULL)
		return -GAMMA_ERR_GENERIC;

/*  We should have collected "num_nodes" participants already!  */
	if (p->vm.how_many_nodes != num_nodes)
		return -GAMMA_ERR_GENERIC;

/*
 *  Collect participant nodes, init the virtual GAMMA, give instance
 *  number 0 to local proc (master).
 */

	p->inst_table[0].whoami = 0;
	p->inst_table[0].has_pkts_in_rx_tmp_ring = 0;

	for (i=0;i<num_nodes;i++) {
		for (j=0;j<MAX_NUM_NODES;j++)
			if (p->vm.wish_flag[j] > 0) {
				p->vm.wish_flag[j]--;
				break;
			}
			else
/*  "wish_flag" is negative if relative to the node of master proc  */
			if (p->vm.wish_flag[j] < 0) {
				p->vm.wish_flag[j]++;
				break;
			}
		if (j>=MAX_NUM_NODES)
			/*  Should never arrive here!  */
			return -GAMMA_ERR_GENERIC;
		p->eth_table[i] = eth_table[j];
		p->vm.node_list[i] = j;
	}

	procnum = 0;
	for (i=0;i<num_nodes;i++) {
		if (p->vm.node_list[i] == phys_whoami) {
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printk("sys_gamma_vm_create_pid():  too many proc instances on this node.  Aborting...\n");
/*
 *  Virtual GAMMA not yet created across the network.  Prevent DESTROY
 *  message from being transmitted by sys_gamma_vm_destroy()...
 */
				p->vm.how_many_nodes = 1;
/*  Destroy the virtual GAMMA  */
				sys_gamma_vm_destroy(par_pid);
				return -GAMMA_ERR_GENERIC;
			}
			p->inst_table[procnum-1].whoami = i;
			p->inst_table[procnum-1].has_pkts_in_rx_tmp_ring = 0;
			for (j=0;j<=MAX_PORT_NUMBER;j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->task = NULL;
				pp->has_rec_handler =
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_same_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = (l2s *)kmalloc(
					HEADER_SIZE+MAX_1ST_PIECE_SIZE,
					GFP_KERNEL
				);
				if (pp->header == NULL) {
printk("sys_gamma_vm_create_pid():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
		}
	}

	p->num_locinst = procnum;

/*  Return the chosen nodes to the caller in user space.  */
	memcpy_tofs(
		user_space_node_list,p->vm.node_list,
		num_nodes*sizeof(unsigned char)
	);

/*  Register the virtual GAMMA across the network, via broadcast  */
	gamma_kernel_send(
		BROADCAST,par_pid,VM_CREATE,
		(void *)(&(p->vm)),sizeof(vm_descriptor),1
	);
/*  Wait enough (1 msec) for the registration to occur remotely  */
	busywait(1000);

	return par_pid;

}


/*
 *  The invoking process informs all nodes that it would need a
 *  virtual GAMMA identified by "par_pid" and expects "num_nodes"
 *  other processes to join the virtual GAMMA as well.
 */
int
sys_gamma_vm_wish
(unsigned long par_pid, unsigned long num_nodes, int first_time)
{

	unsigned long i;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

#ifdef NO_IRQ_RECV
/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
	sys_gamma_poll(par_pid,PROCNUM_NONE);
#endif

	if (first_time)
		if (par_proc_table[par_pid] == (par_entry *)NULL) {
/*  Allocate a descriptor for the new virtual GAMMA  */
			par_proc_table[par_pid] = (par_entry *)
				__get_free_pages(GFP_KERNEL,par_entry_magnitude);
			p = par_proc_table[par_pid];
			if (p == (par_entry *)NULL) {
printk("sys_gamma_vm_wish(): kmalloc() failed\n");
				return -GAMMA_ERR_GENERIC;
			}
			for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
				p->inst_table[i].pid = -1;
				p->inst_table[i].task = NULL;
				p->inst_table[i].whoami = -1;
			}
			p->vm.how_many_nodes  = 0;
			for (i=0;i<MAX_NUM_NODES;i++)
				p->vm.wish_flag[i] = 0;
		}

	p = par_proc_table[par_pid];
	if (p == (par_entry *)NULL)
		return -GAMMA_ERR_GENERIC;

	if (p->vm.how_many_nodes < num_nodes) {
		if (first_time) {
/*
 *  Local proc expresses its wish to join the
 *  virtual GAMMA #par_pid, and tells that to
 *  every node in the cluster.
 */
			p->vm.how_many_nodes++;
			p->vm.wish_flag[phys_whoami]++;
			gamma_kernel_send(BROADCAST,par_pid,VM_WISH,NULL,0,1);
		}
		return -GAMMA_ERR_TRYAGAIN;
	}

/*  All procs have expressed their wish to join the virtual GAMMA  */

/*
 *  Is the local proc master or slave?  The rule is:
 *  The process allocated to the node with lowest physical rank
 *  will be the master.  If more processes are on that node, follow the FCFS
 *  policy (mutual exclusion is required by this routine).
 */
	for (i=0;i<MAX_NUM_NODES;i++) {
		if (!p->vm.wish_flag[i])
			continue;
		if (p->vm.wish_flag[i] < 0)
/*
 *  Master is another local proc instance (why?  See below...)
 *  The caller will be slave.
 */
			return 1;
		if (i == phys_whoami) {
/*
 *  The caller will be master.  Changing the sign to "wish_flag" prevents other
 *  local proc instances to take the same decision.
 */
			p->vm.wish_flag[i] *= -1;
			return 0;
		}
/*  The caller will be slave   */
		return 1;
	}

/*  The caller will be slave   */
	return 1;

}


/*
 *  The invoking process needs the mapping "instance # --> physical node #" .
 *  This functionality is also contained in sys_gamma_vm_create() and
 *  sys_gamma_vm_create_pid().
 *  NOTE:  the argument "user_space_node_list" points to user space!!!
 */
int
sys_gamma_vm_getinfo
(unsigned long par_pid, void *user_space_node_list)
{

/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	if (par_proc_table[par_pid] == NULL)
		return -GAMMA_ERR_GENERIC;

	memcpy_fromfs(
	par_proc_table[par_pid]->vm.node_list,user_space_node_list,
	par_proc_table[par_pid]->vm.how_many_nodes*sizeof(unsigned char)
	);
	return 0;

}


/*
 *  Destroy a virtual GAMMA across the network.
 *  It is invoked only by the master node of that virtual GAMMA, after
 *  the gamma_exit()
 */
int sys_gamma_vm_destroy
(unsigned long par_pid)
{

	unsigned long i,j;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	p = par_proc_table[par_pid];

	if (p == (par_entry *)NULL)
/*  Virtual GAMMA not existing  */
		return -GAMMA_ERR_GENERIC;

	if (p->inst_table[0].whoami != 0)
/*  Not the master node  */
		return -GAMMA_ERR_GENERIC;

	if (p->vm.how_many_nodes > 1)
/*  Destroy the virtual GAMMA across the network  */
		gamma_kernel_send(BROADCAST,par_pid,VM_DESTROY,NULL,0,1);

	for (i=0;i<MAX_NUM_LOCAL_INST;i++)
		if (p->inst_table[i].whoami >= 0)
			for (j=0;j<=MAX_PORT_NUMBER;j++) {
				if (p->inst_table[i].port_table[j].header)
				kfree(p->inst_table[i].port_table[j].header);
			}

	free_pages((unsigned long)p,par_entry_magnitude);
	par_proc_table[par_pid] = (par_entry *)NULL;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  clean up shared ring  */
	spin_lock(&gamma_rxsamenode_lock);
	for (i=0;i<SHARED_RING_SIZE;i++)
		if (shared_ring.desc[i].par_pid == par_pid)
/* par_pid #0 is reserved, rx_gamma_samenode() will eventually discard */
			shared_ring.desc[i].par_pid = 0;
	spin_unlock(&gamma_rxsamenode_lock);
#endif

	return 0;

}


/*
 *  Destroy a virtual GAMMA across the network, and forces all process
 *  instances to exit.
 *  It may be invoked by any process instance (usually upon ctl-C or other
 *  dramatic signals)
 */
int sys_gamma_vm_abort
(unsigned long par_pid)
{

	unsigned long i,j;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	p = par_proc_table[par_pid];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_rx_lock))
		return -GAMMA_ERR_TRYAGAIN;
	if (!spin_trylock(&gamma_rxsamenode_lock)) {
		spin_unlock(&gamma_rx_lock);
		busywait(50*CLOCK);
		return -GAMMA_ERR_TRYAGAIN;
	}
	if (!spin_trylock(&gamma_tx_lock)) {
		spin_unlock(&gamma_rxsamenode_lock);
		spin_unlock(&gamma_rx_lock);
		busywait(50*CLOCK);
		return -GAMMA_ERR_TRYAGAIN;
	}
#endif

	if (p == (par_entry *)NULL) {
/*  Virtual GAMMA #par_pid does not exist  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_rx_lock);
		spin_unlock(&gamma_rxsamenode_lock);
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

	if (p->vm.how_many_nodes > 1) {
/*
 *  Unregister the virtual GAMMA across the network and
 *  require remote process instances to quit.
 */
		gamma_kernel_send(BROADCAST,par_pid,VM_ABORT,NULL,0,0);
/*  Quit _other_ local process instances, if any  */
		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (p->inst_table[i].whoami >= 0)
			if ((p->inst_table[i].task) != current)
			if ((p->inst_table[i].task) != NULL)
/*  When does p->task become NULL???  */
			if ((p->inst_table[i].task)->state != TASK_ZOMBIE)
			if (
			 (p->inst_table[i].task)->pid == p->inst_table[i].pid
			)
				force_sig(SIGSEGV, p->inst_table[i].task);
	}

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  clean up shared ring  */
	for (i=0;i<SHARED_RING_SIZE;i++)
		if (shared_ring.desc[i].par_pid == par_pid)
/* par_pid #0 is reserved, rx_gamma_samenode() will eventually discard */
			shared_ring.desc[i].par_pid = 0;
#endif

	for (i=0;i<MAX_NUM_LOCAL_INST;i++)
		if (p->inst_table[i].whoami >= 0)
			for (j=0;j<=MAX_PORT_NUMBER;j++)
				if (p->inst_table[i].port_table[j].header)
				kfree(p->inst_table[i].port_table[j].header);

	free_pages((unsigned long)p,par_entry_magnitude);
	par_proc_table[par_pid] = (par_entry *)NULL;

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_rx_lock);
	spin_unlock(&gamma_rxsamenode_lock);
	spin_unlock(&gamma_tx_lock);
#endif

	return 0;

}


int
sys_gamma_getconfig
(unsigned long n, void *user_space_eth_table, unsigned long flag)
{

	unsigned long i,j,n_old;


	if (flag) {
		if (total_num_nodes)
			n_old = total_num_nodes;
		else
/*
 *  This might be first invocation after boot (that is, after gamma_start()).
 *  Thus, we have allocated MAX_NUM_NODES Tx tmp rings (presumably more than
 *  needed).
 */
			n_old = MAX_NUM_NODES;
/*  De-allocate Tx tmp ring.  */
		for (i=0;i<n_old;i++)
			for (j=0;j<GAMMA_TX_WIN_SIZE;j++)
				if (tx_tmp_ring[i].buf[j].b != NULL)
					kfree(tx_tmp_ring[i].buf[j].b);
	}

	total_num_nodes = n;

	if (flag) {
/*
 *  Set the initial credit towards each destination.
 *  It may be safer to leave two (or more?) available Rx slots
 *  off-credit for service (non-user) communications.
 */
		for (i=0;i<total_num_nodes;i++) {
			initial_credit[i] = credit_size;
			credit_ceil[i] = credit_size;
			credit_growth_cnt[i] = 0;
			local_credit[i] = initial_credit[i];
			local_credit1[i] = 0;
		}
/*  Clean the Rx tmp ring.  */
		rx_tmp_ring.head = 0;
		rx_tmp_ring.tail = 0;
/*  Allocate and clean the Tx tmp ring.  */
		for (i=0;i<total_num_nodes;i++) {
			for (j=0;j<GAMMA_TX_WIN_SIZE;j++) {
				tx_tmp_ring[i].buf[j].b = (unsigned char *)
					kmalloc(FRAME_SIZ, GFP_KERNEL);
				if (tx_tmp_ring[i].buf[j].b == NULL) {
printk("sys_gamma_getconfig():  kmalloc() failed\n");
					return -1;
				}
				tx_tmp_ring[i].buf[j].l = 0;
			}
			tx_tmp_ring[i].tail = 0;
#ifdef SEQCHECK
			tx_tmp_ring[i].tx_seq = 0;
#endif
		}
#ifdef SEQCHECK
/*  Reset the tables of pkt numbering and congestion flags.  */
		for (i=0;i<total_num_nodes;i++) {
			rx_seq[i] = rx_seq_bcast[i] = 0;
			rx_discard[i] = 0;
			congested[i] = 0;
		}
#endif
	}

/*  get the table of MAC addresses from user space  */
	memcpy_fromfs(
		eth_table,user_space_eth_table,
		MAX_NUM_NODES*sizeof(eth_addr)
	);

DEBUG(
printk("sys_gamma_getconfig():  ETH_TABLE:\n");
for (i=0;i<total_num_nodes;i++)
	printk("%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		eth_table[i].addr[0],
		eth_table[i].addr[1],
		eth_table[i].addr[2],
		eth_table[i].addr[3],
		eth_table[i].addr[4],
		eth_table[i].addr[5]
	);
)

/*  what node is this wrt the physical GAMMA?  */
        for (i=0;i<total_num_nodes;i++) {
                for (j=0;j<6;j++)
                        if (eth_table[i].addr[j] != gamma_device->dev_addr[j])
                                break;
                if (j==6) {
                        phys_whoami = i;
                        break;
                }
        }

	return 0;

}


/*
 *  Registers the PID and the task pointer of the calling process
 *  with the virtual GAMMA specified by par_pid (which is supposed to exist).
 */
int
sys_gamma_register
(unsigned long par_pid)
{

	unsigned long i;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

#ifdef NO_IRQ_RECV
/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
	sys_gamma_poll(par_pid,PROCNUM_NONE);
#endif

	p = par_proc_table[par_pid];

	if (p == (par_entry *)NULL) {
/*  not an existing virtual GAMMA  */
DEBUG_L(printk("sys_gamma_register(): par_pid %ld not created\n",par_pid));
		return -GAMMA_ERR_GENERIC;
	}

DEBUG(
{
unsigned long i;
printk("sys_gamma_register():  par_pid %ld\n", par_pid);
printk("Virtual GAMMA %ld:\nhow_many_nodes %ld\n",
	par_pid,
	p->vm.how_many_nodes);
for (i=0;i<p->vm.how_many_nodes;i++)
	printk("%ld: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		i,
		p->eth_table[i].addr[0],
		p->eth_table[i].addr[1],
		p->eth_table[i].addr[2],
		p->eth_table[i].addr[3],
		p->eth_table[i].addr[4],
		p->eth_table[i].addr[5]
	);
}
);

/*
 *  Since more process instances are allowed on the same station, the
 *  caller has to find the right slot in the virtual GAMMA.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_lock(&gamma_tx_lock);
#endif
	for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
		if (p->inst_table[i].pid != -1)
/*  Another process already registered in this slot  */
			continue;
		if (p->inst_table[i].whoami >= 0) {
/*  A free, valid slot was found  */
			p->inst_table[i].pid = current->pid;
			p->inst_table[i].task = current;
DEBUG(printk(
"Virtual GAMMA %ld: whoami %d\n",par_pid,p->inst_table[i].whoami
));
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return 0;
		}
DEBUG(printk(
"sys_gamma_register(): local host does not belong to par_pid %ld\n", par_pid
));
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}
DEBUG_L(printk("sys_gamma_register(): par_pid %ld:  could not register\n",
par_pid
));
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_GENERIC;

}


int
sys_gamma_register_nospawn
(unsigned long par_pid)
{

	unsigned long i;
	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

#ifdef NO_IRQ_RECV
/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
	sys_gamma_poll(par_pid,PROCNUM_NONE);
#endif

	p = par_proc_table[par_pid];

	if (p == (par_entry *)NULL) {
/*  not an existing virtual GAMMA  */
DEBUG_L(printk(
"sys_gamma_register_nospawn(): par_pid %ld not created\n",par_pid
));
		return -GAMMA_ERR_GENERIC;
	}

DEBUG(
{
unsigned long i;
printk("sys_gamma_register_nospawn():  par_pid %ld, pid %d\n",
	par_pid,
	p->inst_table[0].pid
);
printk("Virtual GAMMA %ld:\nhow_many_nodes %ld\n",
	par_pid,
	p->vm.how_many_nodes);
for (i=0;i<p->vm.how_many_nodes;i++)
	printk("%ld: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		i,
		p->eth_table[i].addr[0],
		p->eth_table[i].addr[1],
		p->eth_table[i].addr[2],
		p->eth_table[i].addr[3],
		p->eth_table[i].addr[4],
		p->eth_table[i].addr[5]
	);
}
);

/*
 *  Since more process instances are allowed on the same station, the
 *  caller has to find the right slot in the virtual GAMMA.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_lock(&gamma_tx_lock);
#endif
	for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
		if (p->inst_table[i].pid != -1)
/*  Another process already registered in this slot  */
			continue;
		if (p->inst_table[i].whoami >= 0) {
/*  A free, valid slot was found  */
DEBUG(printk(
"sys_gamma_register_nospawn(): procnum %ld, whoami %d\n",
i,p->inst_table[i].whoami
));
			p->inst_table[i].pid = current->pid;
			p->inst_table[i].task = current;
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return 0;
		}
		else {
/*
 *  The "whoami" field will eventually be filled up by the master process
 *  when calling sys_gamma_vm_create_pid().  This process will try again
 *  until success.
 */
DEBUG(printk(
"sys_gamma_register_nospawn(): local host does not belong to par_pid %ld yet\n",
par_pid
));
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return -GAMMA_ERR_TRYAGAIN;
		}
	}
DEBUG_L(
printk("sys_gamma_register_nospawn(): par_pid %ld:  could not register\n",
par_pid)
);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_GENERIC;

}


/*
 *  sys_gamma_init, parts 1 & 2 & 3.
 *  These GAMMA primitives look into par_proc_table for the caller's entry
 *  (which was assigned to the caller via a previous invocation to
 *  "gamma_register()")
 *  and then register the caller's tables of receiver buffers and semaphores.
 */

/*
 *  This routine returns the no. of nodes of the virtual GAMMA
 *  which the caller had previously registered to, and
 *  the instance no. of the caller.
 */
int
sys_gamma_init_part1
(
	void *user_space_handler_rec_dispatcher_ptr,
	void *user_space_handler_err_dispatcher_ptr,
	void *user_space_handler_atomic_dispatcher_ptr
)
{

	int found=0,i=0,par_pid;
	par_entry *p;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		if (par_proc_table[par_pid] == (par_entry *)NULL)
			continue;
		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (
				par_proc_table[par_pid]->inst_table[i].pid
				== current->pid
			) {
				found = 1;
				break;
			}
		if (found)
			break;
	}
	if (found) {
		p = par_proc_table[par_pid];
/*
 *  Register the user-space entry points for
 *  rec and err handlers and atomic functions.
 */
		p->inst_table[i].handler_rec_dispatcher =
			user_space_handler_rec_dispatcher_ptr;
		p->inst_table[i].handler_err_dispatcher =
			user_space_handler_err_dispatcher_ptr;
		p->inst_table[i].handler_atomic_dispatcher =
			user_space_handler_atomic_dispatcher_ptr;
	        return (int)(
			((unsigned long)(p->vm.how_many_nodes) << 16) +
			(unsigned long)p->inst_table[i].whoami
		);
	}
	return -GAMMA_ERR_GENERIC;

}


/*
 *  This routine returns the "procnum" of the caller and the par_pid
 *  of the virtual GAMMA which the caller had previously registered to.
 */
int
sys_gamma_init_part2
(
	void *user_space_sem1_table_ptr,
	void *user_space_pre_sem_table_ptr,
	void *user_space_semerr_table_ptr
)
{

	int found=0,i=0,par_pid;
	par_entry *p;

	for (par_pid=1;par_pid<=MAX_PAR_PID;par_pid++) {
		if (par_proc_table[par_pid] == (par_entry *)NULL)
			continue;
		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (
				par_proc_table[par_pid]->inst_table[i].pid
				== current->pid
			) {
				found = 1;
				break;
			}
		if (found)
			break;
	}
	if (found) {
		p = par_proc_table[par_pid];
/*  Register the tables of user-space semaphores.  */
		p->inst_table[i].sem1_table =
			user_space_sem1_table_ptr;
#ifdef USE_DAMNICK
		p->inst_table[i].pre_sem_table =
			user_space_pre_sem_table_ptr;
#endif
		p->inst_table[i].semerr_table =
			user_space_semerr_table_ptr;
/*  Return the "procnum" of the caller.  */
	        return (int)(
			(unsigned long)(par_pid << 16) +
			(unsigned long)i
		);
	}
	return -GAMMA_ERR_GENERIC;

}


/*
 *  This routine returns the clock freq. in Mhz of the local host.
 */
int
sys_gamma_init_part3
(void)
{

	int found=0,i=0,par_pid;

	for (par_pid=1;par_pid<=MAX_PAR_PID;par_pid++) {
		if (par_proc_table[par_pid] == (par_entry *)NULL)
			continue;
		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (
				par_proc_table[par_pid]->inst_table[i].pid
				== current->pid
			) {
				found = 1;
				break;
			}
		if (found)
			break;
	}
	if (found) {
/*  Return the CPU clock freq. in MHz  */
	        return cpu_khz/1000;
	}
	return -GAMMA_ERR_GENERIC;

}


int
sys_gamma_exit
(unsigned long par_pid, unsigned long procnum)
{

	par_entry *p;


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	p = par_proc_table[par_pid];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_rx_lock))
		return -GAMMA_ERR_TRYAGAIN;
	if (!spin_trylock(&gamma_rxsamenode_lock)) {
		spin_unlock(&gamma_rx_lock);
		busywait(50*CLOCK);
		return -GAMMA_ERR_TRYAGAIN;
	}
	if (!spin_trylock(&gamma_tx_lock)) {
		spin_unlock(&gamma_rxsamenode_lock);
		spin_unlock(&gamma_rx_lock);
		busywait(50*CLOCK);
		return -GAMMA_ERR_TRYAGAIN;
	}
#endif

	if (p == (par_entry *)NULL) {
/*  Virtual GAMMA #par_pid does not exist  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_rx_lock);
		spin_unlock(&gamma_rxsamenode_lock);
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

	if (p->inst_table[procnum].pid != current->pid) {
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_rx_lock);
		spin_unlock(&gamma_rxsamenode_lock);
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

/*  unsubscribe from given virtual GAMMA  */
	p->inst_table[procnum].pid = -1;

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_rx_lock);
	spin_unlock(&gamma_rxsamenode_lock);
	spin_unlock(&gamma_tx_lock);
#endif

	return 0;

}
